/******************************************************************************
 *
 * Filename: classfile.h
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Class file datatypes.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#ifndef _classfile_h

#define _classfile_h

#include "types.h"

#define CHECKCONSTANT( cf, index, kind ) ( cf && (index > 0) && (index < cf->constant_pool_count) && cf->constant_pool[index] && (cf->constant_pool[index]->tag == kind) )

/* RHT defined */
#define T_ANY		0
#define T_OBJREF	1
#define T_ARRAY		2
#define T_RETADDR	3

/* newarray defined */
#define T_BOOLEAN	4
#define T_CHAR		5
#define T_FLOAT		6
#define T_DOUBLE	7
#define T_BYTE		8
#define T_SHORT		9
#define T_INT		10
#define T_LONG		11

/* RHT defined */
#define T_DOUBLE1	12
#define T_LONG1		13
#define T_LAST		13


/**************** Constant Pool Entries ****************/

#define CONSTANT_Utf8		1
#define CONSTANT_Integer	3
#define CONSTANT_Float		4
#define CONSTANT_Long		5
#define CONSTANT_Double		6
#define CONSTANT_Class		7
#define CONSTANT_String		8
#define CONSTANT_Fieldref	9
#define CONSTANT_Methodref	10
#define CONSTANT_InterfaceMethodref 11
#define CONSTANT_NameAndType 12


/* Constant Pool Entry Header */

typedef struct _cp_info
{
	u1 tag;		/* Constant type */
	/* u1* info; */
} cp_info;


typedef struct _CONSTANT_Class_info
{
	u1 tag;
	u2 name_index;

	/* extra runtime data */
	struct _ClassFile* cf;
} CONSTANT_Class_info;


typedef struct _CONSTANT_Fieldref_info
{
	u1 tag;
	u2 class_index;
	u2 name_and_type_index;

	/* extra runtime data */
	struct _field_info* info;
} CONSTANT_Fieldref_info;


typedef struct _CONSTANT_Methodref_info
{
	u1 tag;
	u2 class_index;
	u2 name_and_type_index;

	/* extra runtime data */
	struct _method_info* info;
} CONSTANT_Methodref_info;


typedef struct _CONSTANT_InterfaceMethodref_info
{
	u1 tag;
	u2 class_index;
	u2 name_and_type_index;

	/* extra runtime data */
	struct _method_info* info;
} CONSTANT_InterfaceMethodref_info;


typedef struct _CONSTANT_String_info
{
	u1 tag;
	u2 string_index;

	/* extra runtime data */
	struct _ObjRef* objref;
} CONSTANT_String_info;


typedef struct _CONSTANT_Integer_info
{
	u1 tag;
	u4 bytes;
} CONSTANT_Integer_info;


typedef struct _CONSTANT_Float_info
{
	u1 tag;
	u4 bytes;
} CONSTANT_Float_info;


typedef struct _CONSTANT_Long_info
{
	u1 tag;
	u4 high_bytes;
	u4 low_bytes;
} CONSTANT_Long_info;


typedef struct _CONSTANT_Double_info
{
	u1 tag;
	u4 high_bytes;
	u4 low_bytes;
} CONSTANT_Double_info;


typedef struct _CONSTANT_NameAndType_info
{	
	u1 tag;
	u2 name_index;
	u2 descriptor_index;
} CONSTANT_NameAndType_info;


typedef struct _CONSTANT_Utf8_info
{	
	u1 tag;
	u2 length;
	u1* bytes;
} CONSTANT_Utf8_info;


/**************** Attributes ****************/

/* Attribute Header */

typedef struct _attribute_info
{
	u2 attribute_name_index;
	u4 attribute_length;
	/* u1* info; */
} attribute_info;


/* Code attribute */

typedef struct _Exception_table
{
	u2 start_pc;
	u2 end_pc;
	u2 handler_pc;
	u2 catch_type;
} Exception_table;

typedef struct _Code_attribute
{
	u2 attribute_name_index;
	u4 attribute_length;
	u2 max_stack;
	u2 max_locals;
	u4 code_length;
	u1* code;
	u2 exception_table_length;
	Exception_table** exception_table;
	u2 attributes_count;
	attribute_info** attributes;
} Code_attribute;


/* Source File Attribute */

typedef struct _SourceFile_attribute
{
    u2 attribute_name_index;
    u4 attribute_length;
    u2 sourcefile_index;
} SourceFile_attribute;


/* Line Number Table Attribute */

typedef struct _LineNumber_table
{
	u2 start_pc;	     
	u2 line_number;	     
} LineNumber_table;

typedef struct _LineNumberTable_attribute
{
	u2 attribute_name_index;
	u4 attribute_length;
	u2 line_number_table_length;
	LineNumber_table** line_number_table;
} LineNumberTable_attribute;


/* Local Variable Table */

typedef struct _LocalVariable_table
{
	u2 start_pc;
	u2 length;
	u2 name_index;
	u2 descriptor_index;
	u2 index;
} LocalVariable_table;

typedef struct _LocalVariableTable_attribute
{
	u2 attribute_name_index;
	u4 attribute_length;
	u2 local_variable_table_length;
	LocalVariable_table **local_variable_table;
} LocalVariableTable_attribute;


/********************* Class Information *********************/

#define ACC_PUBLIC		 0x0001
#define ACC_PRIVATE		 0x0002	/* not cf */
#define ACC_PROTECTED	 0x0004	/* not cf */
#define ACC_STATIC		 0x0008	/* not cf */
#define ACC_FINAL		 0x0010
#define ACC_SUPER		 0x0020	/* not fi */
#define ACC_SYNCHRONIZED 0x0020	/* mi only */
#define ACC_VOLATILE	 0x0040	/* not cf */
#define ACC_TRANSIENT	 0x0080	/* not cf */
#define ACC_NATIVE		 0x0100	/* mi only */
#define ACC_INTERFACE	 0x0200	/* not fi */
#define ACC_ABSTRACT	 0x0400	/* not fi */


/* Field Information */

typedef struct _field_info
{
	u2 access_flags;
	u2 name_index;
	u2 descriptor_index;
	u2 attributes_count;
	attribute_info** attributes;

	/* extra runtime data */
	struct _ClassFile* cf;
	u4 offset;
} field_info;



/* Native Methods */

struct _Stack;
typedef void (*generic_fn)();
typedef int (*NativeWrapper)(struct _Stack* stp, struct _ObjRef* staticself, generic_fn fn);

typedef struct _NativeMethodStr
{
	const char* class_name;
	const char* name;
	const char* desc;
	NativeWrapper wrapper;
	generic_fn arg;
} NativeMethodStr;


/* Method Information */

typedef struct _method_info
{
	u2 access_flags;
	u2 name_index;
	u2 descriptor_index;
	u2 attributes_count;
	attribute_info** attributes;

	/* extra loadtime info */
	u2 index;

	/* extra runtime data */
	struct _ClassFile* cf;
	u1 args;
	u1 breakpoint;
	Code_attribute* ca;   /* ptr to java code */
	LineNumberTable_attribute* lnta;
	LocalVariableTable_attribute* lvta;
	NativeMethodStr* nativestr;   /* ptr to native method structure */
} method_info;


/********************* Class File *********************/

typedef struct _InterfaceEntry
{
	u2 class_index;

	/* extra runtime data */
	struct _ClassFile* cf;
} InterfaceEntry;


typedef struct _ClassFile
{
	u4 magic;
	u2 minor_version;
	u2 major_version;
	u2 constant_pool_count;
	cp_info** constant_pool;
	u2 access_flags;
	u2 this_class;			/* index into constant_pool */
	u2 super_class;			/* index into constant_pool */
	u2 interfaces_count;	/* direct implements */
	InterfaceEntry** interfaces;
	u2 fields_count;
	field_info** fields;
	u2 methods_count;
	method_info** methods;
	u2 attributes_count;
	attribute_info** attributes;

	/* extra runtime data */
	struct _ClassFile *super;   /* direct super-class */
	u2 virtuals_count;           /* size of virtual method table */
	method_info** virtuals;     /* virtual method table */
	char* name;                 /* human-readable name for debugging */
	SourceFile_attribute* sfa;  /* source filename */
	u1 resolved;                /* resolved flag */
	u1 initialised;             /* static initialisation flag */
	u4 object_size;             /* size of an instance */
	u4 static_size;             /* size of static data */
	u4* static_data;            /* pointer to static data */
	struct _ObjRef* classobj;    /* java.lang.Class object for this class */
} ClassFile;


#endif
